--==============================================================================
-- CERN (BE-CO-HT)
-- Pulse generator with trigger
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-03-01
--
-- version: 1.0
--
-- description:
--    This module generates a variable-width pulse. The width is set using the
--    g_pulse_width generic, given in number of clk_i cycles. Assuming a clk_i
--    period of 8ns, the output pulse width is by default 8*16=128ns.
--
--    The pulse_o signal is generated by multiplexing the input trigger and
--    the internally-generated pulse. When the trigger is detected, it is
--    immediately passed to the output in order to avoid jitter. 
--
--    The trigger then passes through a variable-length glitch filter consisting 
--    of N flip-flops, where N is set via g_glitch_flit_len. Once the trigger 
--    has settled to '1' and the glitch filter is all ones, the internal pulse 
--    signal takes over, extending the input pulse up to g_pulse_width.
--
--    Selection between trigger input and internal pulse signal routing to the
--    output is done by an internal pulse rejection signal, which is set once
--    the internal pulse generation mechanism has started, and cleared when
--    the trigger input has settled to zero. This signal assures that trigger
--    signals no greater than the desired pulse width appear at the input, thus
--    safeguarding the output transformers on the blocking pulse.
--
-- dependencies:
--    none
--
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-03-01   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: - 
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity pulse_generator is
  generic 
  (
    -- Pulse width, in number of clk_i cycles
    g_pulse_width     : natural := 15;
    
    -- Length of glitch filter; the longer the glitch filter, the
    -- longer the input trigger needs to be.
    g_glitch_filt_len : natural := 6
  );
  port
  (
    clk_i   : in  std_logic;
    rst_n_i : in  std_logic;
    
    -- Trigger input, has to be '1' to assure pulse output with delay no greater
    -- than internal gate delays.
    trig_i  : in  std_logic;
    
    -- Pulse output, active-high
    pulse_o : out std_logic
  );
end entity pulse_generator;


architecture behav of pulse_generator is

  --============================================================================
  -- Function and procedure declarations
  --============================================================================
  function f_log2_size (A : natural) return natural is
  begin
    for I in 1 to 64 loop               -- Works for up to 64 bits
      if (2**I >= A) then
        return(I);
      end if;
    end loop;
    return(63);
  end function f_log2_size;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Pulse length counter
  signal width_cnt      : unsigned(f_log2_size(g_pulse_width)-1 downto 0);
  
  -- Pulse-specific signals
  signal pulse          : std_logic;
  signal pulse_reject   : std_logic;
  
  -- Glitch filter
  signal glitch_filt    : std_logic_vector(g_glitch_filt_len downto 0);
  signal glitch_filt_d0 : std_logic;

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Output logic
  --============================================================================
  pulse_o <= trig_i when (pulse_reject = '0') else
             pulse;

  --============================================================================
  -- Glitch filtration logic
  --============================================================================
  glitch_filt(0) <= trig_i;
  gen_glitch_filt: if (g_glitch_filt_len > 0) generate
    p_glitch_filt: process (clk_i)
    begin
      if rising_edge(clk_i) then
        if (rst_n_i = '0') then
          glitch_filt(g_glitch_filt_len downto 1) <= (others => '0');
        else
          glitch_filt(g_glitch_filt_len downto 1) <= glitch_filt(g_glitch_filt_len-1 downto 0);
        end if;
      end if;
    end process p_glitch_filt;
  end generate gen_glitch_filt;

  --============================================================================
  -- Pulse generation logic
  --============================================================================
  p_gen_pulse: process(clk_i)
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        width_cnt      <= (others => '0');
        pulse          <= '0';
        glitch_filt_d0 <= '0';
      else
        -- FF to detect rising edge on glitch filter output
        glitch_filt_d0 <= glitch_filt(g_glitch_filt_len);

        -- Start outputting pulse with controlled length once the glitch filter
        -- has stabilized
        if (glitch_filt = (glitch_filt'range => '1')) and (glitch_filt_d0 = '0') then
          pulse <= '1';
        end if;

        if (pulse = '1') then
          width_cnt <= width_cnt + 1;
          -- Reset pulse length counter and clear output pulse when reached max 
          -- length. The max length is given by the module input, minus the 
          -- glitch filter length (due to the flip-flops the pulse goes through). 
          --
          -- The "-2" is first because the counter starts from zero, thus the max 
          -- length should be pulse_len-1, and second because the pulse is set 
          -- on the clock edge following the glitch filter output settling to all
          -- ones.
          if (width_cnt = g_pulse_width-g_glitch_filt_len-2) then
            width_cnt <= (others => '0');
            pulse     <= '0';
          end if;
        end if;

      end if;
    end if;
  end process p_gen_pulse;

  -- This process is used to reject pulses longer than the pulse width defined
  -- via the generic. This safeguards the blocking output transformers from
  -- reaching saturation current when an INV-TTL signal is applied at the TTL
  -- input without setting the LEVEL switch appropriately. It also guards the
  -- transformers when no signal is present at the TTL input and the LEVEL switch
  -- is set to INV-TTL.
  --
  -- When the pulse signal is set, the pulse_reject signal is also set. The
  -- pulse_reject signal is cleared when the pulse input has settled back to '0'.
  --
  -- By using the pulse_reject signal as an enable for the output, the pulse output 
  -- width is cut to the desired size.
  p_pulse_reject: process(clk_i)
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        pulse_reject <= '1';
      elsif (pulse = '1') then
        pulse_reject <= '1';
      elsif (trig_i = '0') then
        pulse_reject <= '0';
      end if;
    end if;
  end process p_pulse_reject;


end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
